/*
Client, treba znat ip servera kojem salje svoje stanje
Donja ploca, A1
*/

// Ovdje su sacuvani podaci o mrezi te kako se na nju spojiti (radi sigurnosti
// je to bolje drzati u zasebnom file-u).
#include "arduino_secrets.h"
// Sadrzi funkcije za punjenje i citanje poruka
#include "komunikacija.h"
// Popis koja ledica je spojena na koji pin
#include "pinovi_ledice.h"
// Funkcija za stavaranje zvukova
#include "buzzer.h"

// Dodatni library-ji za wifi omoguciti.
#include <WiFi101.h>
#include <WiFiUdp.h>

// Sifre za spajanje na wifi
char ssid[] = SECRET_SSID;
char pass[] = SECRET_PSWD;

unsigned int localPort = LOCAL_PORT;
// Trebamo znati na koji port na server se spajamo
unsigned int serverPort = SERVER_PORT;

// Varijabla za spremanje nase IP adressse na wifiju.
// Neznam jel na tom wifiju mozemo postaviti staticku IP adresu?
IPAddress ip;
// Preko ove varijable cemo slati i primati nase poruke.
WiFiUDP Udp;

char primljena_poruka[255];

// Pinovi joystick
int joystick_x = A0;
int joystick_y = A1;
int joystick_z = A2;

// Granice za ocitanje potenciometara sa joystick-a
int centar = 512;
int deadzone = 256;

// Spremanje pomaka dobivenog od joystick-a za kretanje po poljima.
int pomak_x, pomak_y;

// Trenutni delay izvucen kao varijabla da mozemo ubrzati/usporiti kojom brzinom
// ce se glavna petalja izvoditi
int delay_ms = 300;

constexpr int kBrojBrodova = 3;

// Pamcenje stanja ploca (ovo su 2D liste za stanje ledica - upaljene/ugasene)
bool zelena_stanja[3][3];
bool crvena_stanja[3][3];

// Kako bi prikazali na kojem smo trenutnom polju, moramo pamtiti trenutne
// indekse u listi za x i y smjer kako bi mogli blinkati odredenu ledicu.
int trenutno_polje_x = 0;
int trenutno_polje_y = 0;

bool pobjednik = false;

enum Stanje {
  Inicijalizacija = 0,
  PostavljanjeBrodova = 1,
  CekanjePostavljanjeDrugoga = 2,
  IgramJa = 3,
  IgraDrugi = 4,
  Kraj = 5
} stanje;

// Zastavica za cuvanje jel su postavljali nasi ili protivnicki brodovi.
bool poslao_svoje_brodove = false;
bool protivnik_spreman = false;

void resetStanja() {
  pobjednik = false;

  poslao_svoje_brodove = false;
  protivnik_spreman = false;

  for (int i = 0; i < 3; ++i) {
    for (int j = 0; j < 3; ++j) {
      zelena_stanja[i][j] = 0;
      crvena_stanja[i][j] = 0;
    }
  }
}

void setup() {
  stanje = Inicijalizacija;

  // Otvaranje serial monitora da mozemo gledati poruke koje ispisuje arduino
  // (primarno da znamo sto u nekoj fazi radi).
  Serial.begin(9600);

  resetStanja();

  pinMode(buzzer, OUTPUT);

  // Postavljanje input pinova da mozemo citati joystick
  pinMode(joystick_x, INPUT);
  pinMode(joystick_y, INPUT);
  pinMode(joystick_z, INPUT_PULLUP);

  // Postavljanje svih pinova za ledice na OUTPUT
  for (int i = 0; i < 3; ++i) {
    for (int j = 0; j < 3; ++j) {
      pinMode(led_zeleno[i][j], OUTPUT);
      pinMode(led_crveno[i][j], OUTPUT);
    }
  }

  // Spajanje na wifi mrezu
  int status = WiFi.begin(ssid, pass);
  while (status != WL_CONNECTED) {
    Serial.println("** Can't connect, retry **");
    status = WiFi.begin(ssid, pass);
    // wait 5 seconds for connection:
    delay(5000);
  }

  // Ispis vlastite adrese i otvaranje UDP kanala za slanje poruka.
  Serial.println("You're connected to the network");
  ip = WiFi.localIP();
  Serial.print("Your IP: ");
  Serial.println(ip);
  Serial.println("Starting connection to server...");
  // Ako se sljedeca poruka ispise, znaci da je server vec otvorio ovu vezu i da
  // smo se spojili na njega.
  if (Udp.begin(localPort)) {
    Serial.println("Udp started");
  }

  stanje = PostavljanjeBrodova;
}

// Funkcija s kojom prolazimo po stanjima ledica sacuvanima u led_zeleno i
// led_crveno te palimo/gasimo potrebne ledice.
void paljenje_ledica() {
  for (int i = 0; i < 3; ++i) {
    for (int j = 0; j < 3; ++j) {
      if (zelena_stanja[i][j]) {
        digitalWrite(led_zeleno[i][j], HIGH);
      } else {
        digitalWrite(led_zeleno[i][j], LOW);
      }
      if (crvena_stanja[i][j]) {
        digitalWrite(led_crveno[i][j], HIGH);
      } else {
        digitalWrite(led_crveno[i][j], LOW);
      }
    }
  }
}

// Funckija za brojanje upaljenih zelenih ledica.
int broj_upaljenih_ledica() {
  int broj_upaljenih = 0;
  for (int i = 0; i < 3; ++i) {
    for (int j = 0; j < 3; ++j) {
      if (zelena_stanja[i][j]) {
        broj_upaljenih++;
      }
    }
  }
  return broj_upaljenih;
}

void promjeniStanje(Stanje novo) { stanje = novo; }

// Funkcija za blinkanje jedne specificne ledice (npr. za prikaz joystick-a)
void blinkaj(int led) {
  delay(70);
  digitalWrite(led, HIGH);
  delay(70);
  digitalWrite(led, LOW);
}

void posaljiNaPlocu(IPAddress adresa, char *poruka) {
  Udp.beginPacket(adresa, localPort);
  Udp.write(poruka);
  Udp.endPacket();
}

bool procitajPoruku(char *poruka) {
  int packetSize = Udp.parsePacket();
  if (packetSize) {
    // IPAddress remoteIp = Udp.remoteIP();
    // uint16_t remotePort = Udp.remotePort();

    // read the packet into packetBufffer
    int len = Udp.read(poruka, 255);
    if (len > 0) {
      poruka[len] = 0;
    }
    return true;
  }
  return false;
}

// Stanje se salje svakih nekoliko puta kako se wifi mreža ne bi preopteretila.
int poslo_puta_stanje = 0;
void loop() {

  paljenje_ledica();

  // Odabir izvršavanja faze igre.
  switch (stanje) {
  case PostavljanjeBrodova:
    fazaPostavljanjaBrodova();
    break;
  case CekanjePostavljanjeDrugoga:
    fazaCekanjePostavljanjeDrugoga();
    break;
  case IgramJa:
    fazaIgramJa();
    break;
  case IgraDrugi:
    fazaIgraDrugi();
    break;
  case Kraj:
    fazaKraj();
    break;
  default:
    break;
  };

  poslo_puta_stanje++;
  if (poslo_puta_stanje > 4) {
    char poruka[2];
    poruka[0] = 'S'; // saljemo stanje
    poruka[1] = (char)(stanje + 1);
    posaljiNaPlocu(moja_gornja, poruka);
    poslo_puta_stanje = 0;
  }

  delay(delay_ms);
}

/*
 *
 * Faze igre
 *
 */

void posaljiJoystick(IPAddress adresa) {
  // Pretvaranje podatka tipa int u char za slanje
  char poruka[3];
  poruka[0] = 'j'; // Početni znak za joystick
  poruka[1] = (char)(trenutno_polje_x + 1);
  poruka[2] = (char)(trenutno_polje_y + 1);
  posaljiNaPlocu(adresa, poruka);
}

// Vraća stanje palice/utisnute gljive.
bool joystickKretanje() {
  // Procitaj ulazne pinove od joystick-a
  int reading_x = analogRead(joystick_x);
  int reading_y = analogRead(joystick_y);
  int reading_z = digitalRead(joystick_z);

  // Pretvaramo broj od [0, 1024] na pomak joysticka [-1, 0, 1] za x i y smjer
  if (reading_x < (centar - deadzone))
    pomak_x = -1;
  else if (reading_x > (centar + deadzone))
    pomak_x = 1;
  else
    pomak_x = 0;

  if (reading_y < (centar - deadzone))
    pomak_y = -1;
  else if (reading_y > (centar + deadzone))
    pomak_y = 1;
  else
    pomak_y = 0;

  // Pomak trenutnog polja
  trenutno_polje_x += pomak_x;
  if (trenutno_polje_x < 0)
    trenutno_polje_x = 0;
  if (trenutno_polje_x > 2)
    trenutno_polje_x = 2;
  trenutno_polje_y += pomak_y;
  if (trenutno_polje_y < 0)
    trenutno_polje_y = 0;
  if (trenutno_polje_y > 2)
    trenutno_polje_y = 2;

  // Ocitavanje palice da vidimo promjenimo stanje trenutne ledice.
  // Tipka je HIGH dok nije pritusnuta tipka. Dok je pritisnuta, onda je LOW.
  if (reading_z == HIGH) {
    return false;
  } else {
    return true;
  }
}

void fazaPostavljanjaBrodova() {
  bool utisnuta_gljiva = joystickKretanje();
  if (utisnuta_gljiva) {
    zelena_stanja[trenutno_polje_x][trenutno_polje_y] =
        !zelena_stanja[trenutno_polje_x][trenutno_polje_y];
  }

  // Provjeri jel dovoljno ledica postavljeno da se ide na iducu fazu
  int broj_upaljenih = broj_upaljenih_ledica();
  if (broj_upaljenih == kBrojBrodova) {
    stanje = CekanjePostavljanjeDrugoga;
  }

  // Blinkaj trenutni odabir
  blinkaj(led_crveno[trenutno_polje_x][trenutno_polje_y]);
}

void fazaCekanjePostavljanjeDrugoga() {
  // posalji stanje trenutne ploce protivnickoj gornjoj ploci
  char poruka[4];
  poruka[0] = 'm'; // Saljemo plocu
  serializePloca(zelena_stanja, &poruka[1]);
  posaljiNaPlocu(protivnicka_gornja, poruka);

  // Provjeravaj jel moja i protivnička gornja ploča zna za sve moje brodove,
  // nakon tog nastavljamo.
  bool nesto_procitao = procitajPoruku(primljena_poruka);
  if (nesto_procitao) {
    if (primljena_poruka[0] == 'a') {
      poslao_svoje_brodove = true;
    } else if (primljena_poruka[0] == 'b') {
      protivnik_spreman = true;
    }
  }
  if (poslao_svoje_brodove && protivnik_spreman) {
    // Igrač 0 kreće prvi.
    if (igrac == 0) {
      stanje = IgramJa;
    } else {
      stanje = IgraDrugi;
    }
  }
}

void ispisiJoystick() {
  Serial.print("Joystick: ");
  Serial.print(trenutno_polje_x);
  Serial.print(", ");
  Serial.println(trenutno_polje_y);
}

int poslo_puta_joystick = 0;

void fazaIgramJa() {
  bool utisnuta_gljiva = joystickKretanje();

  poslo_puta_joystick++;
  if (poslo_puta_joystick > 4) {
    posaljiJoystick(moja_gornja);
    poslo_puta_joystick = 0;
  }

  // debug
  // blinkaj(led_crveno[trenutno_polje_x][trenutno_polje_y]);

  if (utisnuta_gljiva) {
    char poruka[3];
    poruka[0] = 'p'; // potop
    poruka[1] = (char)(trenutno_polje_x + 1);
    poruka[2] = (char)(trenutno_polje_y + 1);
    posaljiNaPlocu(moja_gornja, poruka);
  }
  bool nesto_procitao = procitajPoruku(primljena_poruka);
  if (nesto_procitao) {
    if (primljena_poruka[0] == 'p') {
      // Uspjesno je primljen pucanj
      zvuk_pucaj(buzzer);
      stanje = IgraDrugi;
    }
    if (primljena_poruka[0] == 'w') {
      // Primio sam poruku da sam pobjedio.
      zvuk_pobjednicki(buzzer);
      pobjednik = true;
      stanje = Kraj;
    }
  }
}

int broj_upaljenih_ledica(bool stanja[3][3]) {
  int broj_upaljenih = 0;
  for (int i = 0; i < 3; ++i) {
    for (int j = 0; j < 3; ++j) {
      if (stanja[i][j]) {
        broj_upaljenih++;
      }
    }
  }
  return broj_upaljenih;
}

bool novi_pogodeni_provjeri(bool crvena_prije[3][3], bool crvena_poslije[3][3],
                            bool zelena[3][3]) {
  int broj_upaljenih = 0;
  for (int i = 0; i < 3; ++i) {
    for (int j = 0; j < 3; ++j) {
      if (crvena_prije[i][j] != crvena_poslije[i][j]) {
        return zelena[i][j];
      }
    }
  }
  return false;
}

bool crvena_stanja_prije[3][3];
void fazaIgraDrugi() {
  // Primam poruke od protivnicke gornje ploce
  bool nesto_procitao = procitajPoruku(primljena_poruka);
  if (nesto_procitao) {
    if (primljena_poruka[0] == 'm') {
      // Primio sam protivnicka pucanja
      memcpy(crvena_stanja_prije, crvena_stanja, 3 * 3 * sizeof(bool));
      deserializePloca(&primljena_poruka[1], crvena_stanja);
      bool jel_potopljen = novi_pogodeni_provjeri(crvena_stanja_prije,
                                                  crvena_stanja, zelena_stanja);

      if (jel_potopljen) {
        zvuk_pogodak(buzzer);
      } else {
        zvuk_promasaj(buzzer);
      }
      stanje = IgramJa;
    }
    if (primljena_poruka[0] == 'l') {
      // Primio sam poruku sam izgubio
      deserializePloca(&primljena_poruka[1], crvena_stanja);
      zvuk_gubitnicki(buzzer);
      pobjednik = false;
      stanje = Kraj;
    }
    if (primljena_poruka[0] == 'w') {
      // Primio sam poruku da sam pobjedio.
      zvuk_pobjednicki(buzzer);
      pobjednik = true;
      stanje = Kraj;
    }
  }
}

void fazaKraj() {
  char poruka[1];
  if (pobjednik) {
    poruka[0] = 'w'; // pobjednik
  } else {
    poruka[0] = 'l'; // gubitnik
  }
  posaljiNaPlocu(moja_gornja, poruka);

  // Posalji poruku za ponovno igranje svojo gornjoj ploci za prikaz na LCD-u
  bool utisnuta_gljiva = joystickKretanje();
  if (utisnuta_gljiva) {
    char poruka[1];
    poruka[0] = 'R';
    posaljiNaPlocu(moja_gornja, poruka);
    // Pričekaj malo da se ne bi odma resetirala igra
    delay(3000);
  }

  // Potvrdna poruka od gornje ploce da je ona primila poruku za ponavljanje
  // igre. Mogu se prebaciti na stanje PostavljanjeBrodova.
  bool nesto_procitao = procitajPoruku(primljena_poruka);
  if (primljena_poruka[0] == 'R') {
    stanje = PostavljanjeBrodova;
    resetStanja();
  }
}